package aktbus.core;

import aktbus.core.AktMessage;
import java.net.*;
import java.io.*;
import java.util.*;
import org.xml.sax.*;
import java.lang.reflect.Constructor;


/**
* The <em>AktBusReceiver</em> class receives an <em>AktMessage</em> through a socket.
* It can also be subclassed to receive messages which are subclasses of <em>AktMessage</em>.
* To do so, you should create a subclass of <em>AktBusReceiver</em>
* and define the <code>message_creator</code> method which construct your customised message
* from a <em>String</em>.
*
* The original <em>AktBusReceiver</em> receives a single message and disconnects from
* the client by the <code>receive</code> or <code>receive_with_path</code> method.
* In expanding it to accommodate piggy-packed reply, <code>receive_with_extra</code>
* is added, which does not disconnect after receiving an incoming message.
* This allows a customised reply to be sent through the same connection.
* If you use the piggy-packed feature, make sure that you call the
* <code>close_connection</code> method on the <code>AktMessageWithExtra</code> object when
* you finish.
*/
public class AktBusReceiver
{
static final int BUFFER_SIZE=10240;
char buffer[]=new char[BUFFER_SIZE];
ServerSocket server_socket;


/**
* Creates an <em>AktBusReceiver</em> object that listens to a free port.
* To find out the port being chosen, use the <code>get_port()</code> method.
*
* @throws	IOException	When any I/O error occurs.
*/
public AktBusReceiver()
	throws java.io.IOException
{
this(0);
} //end constructor

/**
* Creates an <em>AktBusReceiver</em> object that listens to a certain port.
*
* @param port	The port number to listen.
		If <code>port</code> is <code>0</code>, then a free port will be used.
* @throws	IOException	When any I/O error occurs.
*/
public AktBusReceiver(int port)
	throws java.io.IOException
{
server_socket=new ServerSocket(port);			//create server socket
} //end constructor

/**
* Returns the local port number that the <em>AktBusReceiver</em> is listening to.
*
* @return The local port number in use.
*/
public int get_port()
{
return(server_socket.getLocalPort());
} //end method get_port


/**
* Shutdown the receiver and release the port in use.
*
* @throws java.io.IOException	When an I/O error occurs during closing.
*/
public void close()
	throws java.io.IOException
{
server_socket.close();
} //end method close


/**
* Receive and return an <em>AktMessage</em>.
* This method blocks until a reply is received.
* To receive with timeout, use <code>receive_message(int)</code> instead
* To reply an instance of a subclass of <code>AktMessage</code>,
* use <code>receive_message(int,java.lang.String)</code>
* or <code>receive_message(int,java.lang.Class)</code>.
*
* @return	An instance of the <em>AktMessage</em> class or one of its sub-classes.
* @throws org.xml.sax.SAXException	When the received content cannot be parsed into XML.
* @throws java.io.IOException		If anything goes wrong witht the I/O.
* @throws java.lang.Exception		When other general error occurs.
*/
public AktMessage receive_message()
	throws	org.xml.sax.SAXException,
		java.io.IOException,
		java.lang.Exception
{
return(receive_message(0,java.lang.Class.forName("aktbus.core.AktMessage")));
}

/**
* Receive and return an <em>AktMessage</em> with timeout and expect a reply
* of a certain class.
*
* @param	timeout		The timeout value in second.
* @param	reply_class	The expect class of the replying message in a java.lang.Class object.
*
* @return	An instance of the <em>AktMessage</em> class or one of its sub-classes.
* @throws org.xml.sax.SAXException	When the received content cannot be parsed into XML.
* @throws java.io.IOException		If anything goes wrong witht the I/O.
* @throws java.lang.Exception		When other general error occurs.
*/
public AktMessage receive_message(int timeout,java.lang.Class reply_class)
	throws	org.xml.sax.SAXException,
		java.io.IOException,
		java.lang.Exception
{
return(receive(timeout,reply_class).message);
} //end method receive_message


/**
* Waits and receive an <em>AktMessageWithPath</em>.
*
* @throws org.xml.sax.SAXException	When the received content cannot be parsed into XML.
* @throws java.io.IOException		If anything goes wrong witht the I/O.
* @throws java.lang.Exception		When other general error occurs.
*
* @return	Returns an <em>AktMessageWithPath</em> object.
*/
public AktMessageWithPath receive()
	throws	java.io.IOException,
		org.xml.sax.SAXException,
		java.lang.Exception
{
return(receive(0,"aktbus.core.AktMessage"));
} //end method recieve



/**
* Waits and receive an <em>AktMessageWithPath</em> with timeout.
*
* @param timeout	The timeout value in second.
* @param class_string	The expected class of the reply as a String.
*
* @throw org.xml.sax.SAXException	When the received content cannot be parsed into XML.
* @throw java.io.IOException		If anything goes wrong witht the I/O.
* @throw java.lang.Exception		When other general error occurs.
*
* @return	Returns an AktMessageWithPath.
*/
public AktMessageWithPath receive(int timeout,String class_string)
	throws	java.io.IOException,
		org.xml.sax.SAXException,
		java.lang.Exception
{
return(receive(timeout,java.lang.Class.forName(class_string)));
}

/**
* Waits and receive an <em>AktMessageWithPath</em> with timeout.
*
* @param timeout	The timeout value in second.
* @param result_class	The expected class of the reply as an java.lang.Class object.
*
* @throw org.xml.sax.SAXException	When the received content cannot be parsed into XML.
* @throw java.io.IOException		If anything goes wrong witht the I/O.
* @throw java.lang.Exception		When other general error occurs.
*
* @return	Returns an AktMessageWithPath.
*/
public AktMessageWithPath receive(int timeout,java.lang.Class result_class)
	throws	java.io.IOException,
		org.xml.sax.SAXException,
		java.lang.Exception
{
Socket client_socket;
BufferedReader in;
PrintWriter out;
AktMessageWithExtra extra;		//temp result w/ extra info
AktMessageWithPath result;		//result to be returned

if (timeout<0)
	timeout=0;
if (result_class==null)
	{
	result_class=java.lang.Class.forName("aktbus.core.AktMessage");
	}

//
// accept incoming connection
// if exception happens, it will be propagated to the caller
//
client_socket=server_socket.accept();								//wait for a connection
client_socket.setSoTimeout(timeout);								//set timeout value
in=new BufferedReader(new InputStreamReader(client_socket.getInputStream()),BUFFER_SIZE);	//get input stream from connection
out=new PrintWriter(client_socket.getOutputStream(),true);					//get output stream
extra=receive_from_stream(client_socket,in,out,result_class);					//receive message from stream
in.close();						//close streams
out.close();
client_socket.close();					//close incoming connection
//
// prepare result
//
result=new AktMessageWithPath();			//create AktMessageWithPath object
result.message=extra.message;				//assign field values
result.path=extra.path;
return(result);						//return result
} //end method receive



/**
* Receives a request from the streams and returns an <em>AktMessageWithExtra</em> object.
* The I/O streams and socket are not closed, allowing the receiver to send an optional
* piggy-packed reply.
*
* @param	The input stream.
* @param	The output stream.
* @return	An <em>AktMessageWithExtra</em> object containting the received <em>AktMessage</em>
*		and other extra information.
* @throws org.xml.sax.SAXException	When the received content cannot be parsed into XML.
* @throws java.io.IOException		If anything goes wrong witht the I/O.
* @throws java.lang.Exception		When other general error occurs.
*
*/
private AktMessageWithExtra receive_from_stream(Socket socket,BufferedReader in,PrintWriter out,java.lang.Class result_class)
	throws	java.io.IOException,
		org.xml.sax.SAXException,
		java.lang.Exception
{
AktMessageWithExtra result;		//result
StringTokenizer tokenizer;		//tokenizer for parsing
String line,method_string,uri_string,version_string,body;
int message_length=-1;			//negative means not found in message
String field_name;
int status_code=0;			//status-code to be returned
String status_reason=null;		//reason, if any


		
//
// read & process request line
// if anything goes wrong, we'll have to trap it and send a proper HTTP response
//
try	{
	line=in.readLine();								//read request line
	tokenizer=new StringTokenizer(line);						//create tokeniser
	method_string=new String(tokenizer.nextToken());				//get 1st token, which is the method
	uri_string=new String(tokenizer.nextToken());					//get 2nd token, i.e. URL
	version_string=new String(tokenizer.nextToken());				//get 3rd token, i.e. HTTP version
	if (!method_string.equalsIgnoreCase("POST"))					//check for POST
		{
		status_code=405;							//405: method not allowed
		status_reason="only POST is supported";
		}
	else	//
		//continue if POST is found
		//
		{
		//
		// read & process header line, look for "content-length"
		//
		while ((line=in.readLine())!=null)						//read header until getting a blank line
			{
			if (line.length()==0)							//stop if it is a blank line
				break;
			try	{
				tokenizer=new StringTokenizer(line);				//tokenise string
				field_name=new String(tokenizer.nextToken());			//get 1st token, i.e. the field w/ ":"
				if (field_name.equalsIgnoreCase("content-length:"))		//look for the "content-length" field w/ ":"
					{
					message_length=Integer.parseInt(new String(tokenizer.nextToken()));	//get content-length value
					}
				} catch (NoSuchElementException e)
					{
					status_code=400;
					status_reason="error in parsing headers";
					throw new NoSuchElementException();
					}
			}
		//
		// *** NOTE:	I doubt if we should check for a non-zero-length message here.
		//		Do we allow an empty ATK message?
		//
		//		KH
		//
		//
		// get POST body
		//
		//
		// if message_lenght is negative, that means no Content-Lenght is found in the header
		// we just read until EOF
		//
		if (message_length<0)
			{
			status_code=411;							//411: length required
			status_reason="Content-Length required";				//need content-length
			}
		//
		// if message_length is >=0, then we read up to a certain length
		//
		else	{
			//
			//RMF reports that Unix's Java has a problem in reading from a BufferedReader when
			//the data size exceeds a certain limit (~20000 bytes).
			//This maybe due to the fact that the default buffer size is not big enough.
			//
			//Here I use a quick solution by reading in chunks.
			//I made this 10240 (10K bytes), hoping that it will fit into the memory page boundaries (if any).
			//
			//An alternative solution is to construct a AktMessage from InputSource.
			//This works as far as the sender closes the stream when it finishes sending the XML.
			//
			//K.Hui
			//
			int to_read,byte_read;

			to_read=message_length;						//no. of bytes to read
			body="";							//initiall empty
			while (to_read>0)						//still something to read
				{
				if (to_read>BUFFER_SIZE)
					byte_read=in.read(buffer,0,BUFFER_SIZE);
				else byte_read=in.read(buffer,0,to_read);
				if (byte_read<0) break;					//stop if hitting EOF
				to_read-=byte_read;					//no. to read decreases
				body+=new String(buffer,0,byte_read);			//append to string read
				}
			//
			// now parse & create AktMessage from content read
			//
			result=new AktMessageWithExtra();				//create AktMessageWithpath
			result.path=uri_string;						//put in path
//			result.message=message_creator(body);
			//
			// call constructor of the desired class to create message object from a String
			//
			Class[] arg_classes=new Class[1];				//a class array of 1 element
			Object[] args=new Object[1];					//an object array of 1 element
			arg_classes[0]=java.lang.Class.forName("java.lang.String");	//the java.lang.String class object
			args[0]=body;							//put in the arg
			Constructor c=result_class.getConstructor(arg_classes);		//get the constructor w/ 1 String arg
			try	{
				result.message=(AktMessage)c.newInstance(args);			//call the constructor
				} catch (java.lang.reflect.InvocationTargetException e)
					{
					throw (Exception)e.getTargetException();
					}
			result.in_stream=in;
			result.out_stream=out;						//output stream to piggy-pack reply
			result.socket=socket;
			//
			// everything goes fine! send 200 OK
			//
			out.println("HTTP/1.0 200 OK");				//send HTTP reply status
			out.println("");
			return(result);						//return created AktMessage
			}
		}
	}	//
		// if something goes wrong
		//
	 catch (NoSuchElementException e)				//no more token in header line
		{
		status_code=400;					//400: bad request
		status_reason="error in parsing HTTP request line";
		}
	catch (SAXException e)						//can't parse message content
		{
		status_code=400;					//400: bad request
		status_reason="error in parsing message content";
		}
	catch (IOException e)
		{
		status_code=400;
		status_reason="I/O error";
		}
//
// send HTTP reply status, these are all errors
//
out.println("HTTP/1.0 "+status_code+" "+status_reason);	//send HTTP reply status
out.println();
out.flush();
throw new Exception(status_reason);	//throw Exception to report error, no need to return anything
} //end method receive_from_stream


/**
* Waits and receive an <em>AktMessageWithExtra</em>.
* This is usually used when a piggy-packed reply is needed.
*
* @return	An <em>AktMessageWithExtra</em> object containing the received <em>AktMessage</em> and
*		other extra information.
* @throw org.xml.sax.SAXException	When the received content cannot be parsed into XML.
* @throw java.io.IOException		If anything goes wrong witht the I/O.
* @throw java.lang.Exception		When other general error occurs.
*
* @return	Returns an AktMessageWithPath.
*/
public AktMessageWithExtra receive_with_extra()
	throws	org.xml.sax.SAXException,
		java.io.IOException,
		java.lang.Exception
{
return(receive_with_extra(0,"aktbus.core.AktMessage"));
} //end method receive_with_extra

/**
* Waits and receive an <em>AktMessageWithExtra</em> with timeout.
* This is usually used when a piggy-packed reply is needed.
*
* @param timeout	The timeout value in second.
* @param class_string	The expected class of the reply message as a String.
*
* @return	An <em>AktMessageWithExtra</em> object containing the received <em>AktMessage</em> and
*		other extra information.
* @throw org.xml.sax.SAXException	When the received content cannot be parsed into XML.
* @throw java.io.IOException		If anything goes wrong witht the I/O.
* @throw java.lang.Exception		When other general error occurs.
*
* @return	Returns an AktMessageWithPath.
*/
public AktMessageWithExtra receive_with_extra(int timeout,String class_string)
	throws	java.io.IOException,
		org.xml.sax.SAXException,
		java.lang.Exception
{
return(receive_with_extra(timeout,java.lang.Class.forName(class_string)));
}


/**
* Waits and receive an <em>AktMessageWithExtra</em> with timeout.
* This is usually used when a piggy-packed reply is needed.
*
* @param timeout	The timeout value in second.
* @param result_class	The expected class of the reply message as a java.lang.Class object.
*
* @return	An <em>AktMessageWithExtra</em> object containing the received <em>AktMessage</em> and
*		other extra information.
* @throw org.xml.sax.SAXException	When the received content cannot be parsed into XML.
* @throw java.io.IOException		If anything goes wrong witht the I/O.
* @throw java.lang.Exception		When other general error occurs.
*
* @return	Returns an AktMessageWithPath.
*/
public AktMessageWithExtra receive_with_extra(int timeout,java.lang.Class result_class)
	throws	java.io.IOException,
		org.xml.sax.SAXException,
		java.lang.Exception
{
Socket client_socket;
BufferedReader in;
PrintWriter out;
AktMessageWithExtra result;

if (timeout<0)
	timeout=0;
if (result_class==null)
	result_class=java.lang.Class.forName("aktbus.core.AktMessage");
//
// accept incoming connection
// if exception happens, it will be propagated to the caller
//
client_socket=server_socket.accept();						//wait for a connection
client_socket.setSoTimeout(timeout);						//set timeout value
in=new BufferedReader(new InputStreamReader(client_socket.getInputStream()),BUFFER_SIZE);	//get input stream from connection
out=new PrintWriter(client_socket.getOutputStream(),true);			//get output stream

result=receive_from_stream(client_socket,in,out,result_class);	//receive, keeping streams open
//in.close();				//close input stream
//out.println("what the hack");
//client_socket.close();			//close incoming connection
return(result);				//return result
} //end method receive_with_extra


/**
* The message creating method that constructs a message object from a <em>String</em>.
* When you extend <em>AktBusReceiver</em> to receive customised messages of your own
* class (usually subclass of <em>AktMessage</em>), override this method so that it creates
* messages of your customised class.
* Remember that you'll need to cast the returned object from <em>AktMessage</em>
* to your new subclass.
*
* @param body	A String to be parsed into a message.
* @throws	org.xml.sax.SAXException	When the string cannot be parsed successfully.
* @throws	java.lang.Exception		If other general error occurs.
* @return	An instance of the <em>AktMessage</em> class.
*/
/*
protected AktMessage message_creator(String body)
	throws	org.xml.sax.SAXException,
		java.lang.Exception
{
return new AktMessage(body);
} //end method message_creator
*/

} //end class AktBusReceiver
