package aktbus.core;

import com.sun.xml.tree.*;
import org.xml.sax.*;
import org.w3c.dom.*;
import java.util.*;
import java.net.*;
import java.io.*;
import java.lang.reflect.Constructor;
import sunlabs.brazil.util.http.*;

/**
* The <em>AKTMessage</em> class represents an AKT message according to the AKT-Bus white paper.
* It provides methods to compose and manipulate AKT messages, including sending
* an AKT message to a specific URL.
*/

public class AktMessage extends XmlDocument
{
static String HTTP_VERSION="1.1";	//HTTP version string
protected HashMap name_space=null;		//the map of XML name space definition
static final int BUFFER_SIZE=10240;					//buffer size

//
// constructors
//

/**
*
* Creates an empty AktMessage.
*
*/
public AktMessage()
{
super();				//call superclass (XMLDocument)'s constructor
initialise_namespace();
} //end constructor


/**
* Constructs an <em>AktMessage</em> by parsing an XML string.
* The string should contain the message header and body.
*
* @param message_to_parse The string to be parsed.
*
* @throws SAXException	When a parse error is encounted.
* @throws Exception	When a general exception is found.
*/
public AktMessage(String message_to_parse) throws org.xml.sax.SAXException, Exception
{
this(new InputSource(new StringReader(message_to_parse)));
} //end constructor

/**
* Constructs an <em>AktMessage</em> by reading from an <em>org.xml.sax.InputSource</em>.
*
* @param source	<em>InputSource</em> to read from.
* @throws org.xml.sax.SAXException	When a parsing error occurs.
* @throws java.lang.Exception		When other general erros occur.
*/
public AktMessage(org.xml.sax.InputSource source)
	throws	org.xml.sax.SAXException,
		java.lang.Exception
{
XmlDocument x;
Node root;

try	{
	x=XmlDocument.createXmlDocument(source,false);	//parse string as XML
	root=x.getDocumentElement();		//get root element of XML document
	this.changeNodeOwner(root);		//change its owner to current (AktMessage) object
	this.appendChild(root);			//append nodes (tree) to current empty document
	initialise_namespace();
	} catch (IOException e)
		{
		throw new Exception("ERROR: internal error converting string input source");
		}
} //end constructor




/**
* A general constructor that creates an <em>AktMessage</em>.
*
* When <code>parse_content</code> is false, <code>content</code> is treated as plain text without any XML parsing.
* In this case it is put inside the default XML element </code><aktbus:SimpleContent></code>.
*
* When <code>parse_content</code> is true, the string <code>content</code> is parsed to form
* an XML subtree which will be inserted as the content of the message.
*
* @param sender		The URL of the sender.
* @param receiver	The URL of the receiver.
* @param parse_content	A boolean flag indicating if the content is to be parsed as XML.
* @param content	The content of the message.
* @throws org.xml.sax.SAXException	When a parsing error occurs.
* @throws java.lang.Exception		When other general error occur.
*/
public AktMessage(URL sender,URL receiver,boolean parse_content,String content)
	throws org.xml.sax.SAXException,
	java.lang.Exception
{
initialise_namespace();
if (parse_content)
	//
	// parse content string as XML and create subtree
	//
	{
	XmlDocument x;
	Node content_root;

	try	{
		x=XmlDocument.createXmlDocument(new InputSource(new StringReader(content)), false);	//parse content as XML
		content_root=x.getDocumentElement();				//get root element of XML document
		this.changeNodeOwner(content_root);				//change owner to current object
		this.appendChild(createDOM(this,sender,receiver,content_root));
		} catch (IOException e)
			{
			throw new Exception("ERROR: internal error reading from String");
			}
	}
else	//
	// take content as a plain string inside <aktbus:SimpleContent>
	{
	//
	// Create a element and text node for the message
	//
	ElementNode simple_content, has_simple_message;
	Text message_text;

	simple_content = (ElementNode) this.createElement("aktbus:SimpleContent");		//create aktbus:SimpleContent node
	has_simple_message = (ElementNode) this.createElement("aktbus:hasSimpleMessage");	//create aktbus:hasSimpleMessage node
	has_simple_message.setAttribute("rdf:parseType", "Literal");
	message_text = this.createTextNode(content);						//content as a Literal, no parsing

	// Put the nodes together and add the text
	simple_content.appendChild(has_simple_message);		//simple_content --> has_simple_message
	has_simple_message.appendChild(message_text);		//has_simple_message --> message_text

	// Create the entire AKT document model using the message node created above and
	// apend the whole thing to create the final document.
	this.appendChild(createDOM(this,sender,receiver,simple_content));
	}
} //end constructor


/**
* Creates an <em>AktMessage</em> using the <em>sender</em> and <em>reciever URIs</em> and
* appends the content node as the content of the message.
*
* @param sender The URI of the sender of the message.
* @param receiver The URI of the receiver of the message.
* @param content The Node (subtree) to be the content of the message.
*
* @throws MalformedURLException	When either URI is invalid.
* @throws java.lang.Exception	On other general errors.
*/
public AktMessage(URL sender,URL receiver,org.w3c.dom.Node content)
	throws	java.net.MalformedURLException,
		java.lang.Exception
{
this.appendChild(createDOM(this,sender,receiver,content));
initialise_namespace();
} //end constructor


/**
* Creates an <em>AktMessage</em> from a <em>String</em>.
*
* @param sender		The sender of the <em>AktMessage</em>.
* @param receiver	The receiver of the <em>AktMessage</em>.
* @param simple_content	A <em>String</em> forming the simple content of the <em>AktMessage</em>.
*
*/
public AktMessage(URL sender,URL receiver,String simple_content)
	throws org.xml.sax.SAXException,Exception
{
this(sender,receiver,false,simple_content);
} //end constructor


/**
* Creates an <em>AktMessage</em> with <code>&lt;aktbus:SimpleContent&gt;</code>,
* using <em>String</em> for URLs.
*
* @param sender		The sender URL in a String.
* @param receiver	The receiver URL in a String.
* @param simple_content	The message content in a String.
*
* @throws java.net.MalformedURLExcpetion	When either sender/reciver URI is not valid.
* @throws org.xml.sax.SAXException		When <code>simple_content</code> cannot be parsed as XML.
* @throws java.lang.Exception			On other general errors.
*/
public AktMessage(String sender,String receiver,String simple_content)
	throws	java.net.MalformedURLException,
		org.xml.sax.SAXException,
		java.lang.Exception
{
this(new URL(sender),new URL(receiver),simple_content);
} //end constructor


/**
* Creates an <em>AktMessage</em> as an <em>XmlDocument</em>, with proper header information
* and <em>Node</em> content as the message content.
* When you create an <em>AktMessage</em> with a customised content,
* you can use this method to add an XML subtree into an <em>AktMessage</em> envelope.
*
* @param parent The parent <em>XmlDocument</em> object.
* @param sender The sender of this <em>AktMessage</em>.
* @param receiver The receiver of this <em>AktMessage</em>.
* @param content A <em>Node</em> containing the subtree as the <em>AktMessage</em> content.
*
* @throws java.net.MalformedURLExcpetion	When either sender/reciver URI is not valid.
* @throws java.lang.Exception			On other general errors.
*/
public org.w3c.dom.Node createDOM(com.sun.xml.tree.XmlDocument parent,URL sender,URL receiver,org.w3c.dom.Node content)
	throws	java.net.MalformedURLException,
		java.lang.Exception
{
ElementNode rdf;
ElementNode container, has_simple_header, simple_header;
ElementNode has_sender, has_receiver, date, time;
ElementNode has_content;
Text date_text, time_text, sender_text, receiver_text;
Calendar now;
String date_string, time_string;
int year, month, day;
int hour, min, sec;
String id_string, something;




//build all nodes

rdf = (ElementNode) parent.createElement("rdf:RDF");
container = (ElementNode) parent.createElement("aktbus:Container");
has_simple_header = (ElementNode) parent.createElement("aktbus:hasHeader");
simple_header = (ElementNode) parent.createElement("aktbus:SimpleHeader");
date = (ElementNode) parent.createElement("aktbus:hasDateString");
time = (ElementNode) parent.createElement("aktbus:hasTimeString");
has_content = (ElementNode) parent.createElement("aktbus:hasContent");
now = Calendar.getInstance(); //create current time object
year = now.get(Calendar.YEAR); //get date & time as integers
month = map_month(now.get(Calendar.MONTH));				//get month from Calender & map it into 1-12
day = now.get(Calendar.DAY_OF_MONTH);
hour = now.get(Calendar.HOUR_OF_DAY);
min = now.get(Calendar.MINUTE);
sec = now.get(Calendar.SECOND);
date_string = year + "-" + month + "-" + day; //construct date & time strings
time_string = hour + ":" + min + ":" + sec;
date_text = parent.createTextNode(year + "-" + month + "-" + day); //create XML TextNodes
time_text = parent.createTextNode(hour + ":" + min + ":" + sec);

//build node hierarchy

rdf.appendChild(container);
container.appendChild(has_simple_header);
container.appendChild(has_content);
has_simple_header.appendChild(simple_header);
simple_header.appendChild(date);
simple_header.appendChild(time);
date.appendChild(date_text);
time.appendChild(time_text);

//
// if sender is not null, put in its subtree
//
if (sender!=null)
	{
	//
	// check for HTTP in sender URI
	//
	if (!sender.getProtocol().equalsIgnoreCase("http"))
		throw new MalformedURLException("ERROR: only HTTP is supported");
	//
	// create nodes & subtree
	//
	has_sender = (ElementNode) parent.createElement("aktbus:hasSenderURI");
	sender_text = parent.createTextNode(sender.toString());
	simple_header.appendChild(has_sender);
	has_sender.appendChild(sender_text);
	}

//
// if receiver is not null, put in its subtree
//
if (receiver!=null)
	{
	//
	// check for HTTP in receiver URI
	//
	if (!receiver.getProtocol().equalsIgnoreCase("http"))
		throw new MalformedURLException("ERROR: only HTTP is supported");
	//
	// create nodes & subtree
	//
	has_receiver = (ElementNode) parent.createElement("aktbus:hasReceiverURI");
	receiver_text = parent.createTextNode(receiver.toString());
	has_receiver.appendChild(receiver_text);
	simple_header.appendChild(has_receiver);
	}


// the content to add.
has_content.appendChild(content);

//set rdf:RDF element attributes of name-spaces
try	{
	Iterator key_iterator=this.name_space.keySet().iterator();
	while (key_iterator.hasNext())
		{
		String key=(String)key_iterator.next();			//get key
		String value=(String)this.name_space.get(key);		//get value from key
		rdf.setAttribute(key,value);				//set attribute
		}
	} catch (java.lang.NullPointerException e)
		{
		throw new Exception("Internal error: XML name space not defined");
		}

//set container ID (of message)
something = "Something";
//id_string = something;
//
// RDF Parser doesn't like this...
// KH: how about this?
//
// ***	OK. We finally changed the rdf:ID attribute to aktbus:id.
//	KH 31/07/01
//
if (sender!=null)
	id_string = sender.toString() + "/" + date_string + "/" + time_string + "/" + something;
else id_string = "piggypacked/" + date_string + "/" + time_string + "/" + something;
//	id_string = sender.toString().replace('/','-') +
//		date_string + "-" + time_string + "-" + something;
//	container.setAttribute("rdf:ID", id_string);
container.setAttribute("aktbus:id", id_string);

return(rdf);
} //end method createDOM


//--------------------------------------
// methods to manipulate message content
//--------------------------------------

/**
* Given the tag, find an <em>ElementNode</em> in an <em>AktMessage</em> object by depth-first-search.
* To get the value of a certain element, first search for the <em>ElementNode</em>
* by specifying the tag name.
* Once the <em>ElementNode</em> is found, use <code>get_first_child_node()</code> to
* look for its child which is a text node.
* Then get the value of the text node by <code>getNodeValue()</code>.
*
* @param tagname The name of the node to find.
* @return The node with the text if found, <code>null</code> otherwise.
*/
public com.sun.xml.tree.ElementNode find_element_node(String tagname)
{
Node node;

node=this.getDocumentElement();			//find root node of this XML document

if (node==null)					//if no node is found in document
	return (null);				//return null

if (node.getNodeType()==Node.ELEMENT_NODE)	//if it is an ElementNode, proceed to search
	return(this.find_element_node((ElementNode)node,tagname));
return(null);
} //end method find_element_node


/**
* Search for an <em>ElementNode</em> with a given tag-name under a subtree by depth-first-search.
*
* @param root		The root of the subtree to search
* @param tagname	The name of the node to find.
* @return ElementNode The node with the text if found, <code>null</code> otherwise.
*/
public com.sun.xml.tree.ElementNode find_element_node(com.sun.xml.tree.ElementNode root, String tagname)
{
Node child;
ElementNode result;

if (root==null)					//not found
	return (null);
if (root.getNodeName().equals(tagname))		//this is the one
	return (root);
child=root.getFirstChild();			//otherwise, get 1st child of this node
while (child!=null)
	{
	if (child.getNodeType()==Node.ELEMENT_NODE)
		//if child is an ElementNode
		{
		result=find_element_node((ElementNode)child,tagname);		//recursively search in child's subtree
		if (result!=null)
			return(result);						//found
		}
	child=child.getNextSibling(); //try next sibling
	}
return(null);
} //end method find_element_node


/**
* Return the subtree under the given node as a <em>String</em>.
* This is useful for extracting the content as text for display.
* References to entities will not be translated.
* Notice that it does not include the tag name of the top node.
*
* @param n	The node under which the text node of the subtree is to be extracted.
* @return	A <em>String</em> which contains value of all text node in the subtree.
*/
public String get_subtree_text(org.w3c.dom.Node n)
{
return(get_subtree_text(n,false));
} //end method get_subtree_text


/**
* Return the subtree under the given node as a <em>String</em> with optional
* entity references translation.
* Notice that it does not include the tag name of the top node.
*
* @param n		The node under which the text node of the subtree is to be extracted.
* @param translate	A boolean flag specifying if entity reference translation is needed.
* @return		A <em>String</em> which contains value of all text node in the subtree.
*/
public String get_subtree_text(org.w3c.dom.Node n,boolean translate)
{
StringWriter sw=new StringWriter();

try	{
	n=n.getFirstChild();		//get 1st child
	while (n!=null)
		{
		ugly_write_tree(sw,n,translate);	//append output into StringWriter
		n=n.getNextSibling();
		}
	sw.close();
	} catch (java.io.IOException e)
		{
		System.err.println(e.toString());
		return(null);
		}
return(sw.toString());
} //end method get_subtree_text


/**
* Get the first child node of a specific type under the current <em>ElementNode</em>.
*
* @param parent	The parent node under which we search for its children.
* @param type_wanted	The type of node we are looking for.
*			It is defined in <em>org.w3c.dom.Node</em>.
*			Some of the valid values are:
*			<code>ELEMENT_NODE</code>,
*			<code>ATTRIBUTE_NODE</code>,
*			<code>TEXT_NODE</code>.
* @throws java.lang.Exception	When an error occurs.
*/
public org.w3c.dom.Node get_first_child_node(com.sun.xml.tree.ElementNode parent,int type_wanted)
	throws java.lang.Exception
{
return(get_first_child_node(parent,null,type_wanted));
} //end method get_first_child_node


/**
* Get the first child node matching an optional node_name & type.
*
* @param parent		The Node under which we look for the child.
* @param node_name	An optional node name to match. Specify <code>null</code> to ignore it.
* @param type_wanted	The node type to match.
*
* @return	The child node found. <code>null</code> if not found.
* @throws java.lang.Exception	When an error occurs.
*/
public org.w3c.dom.Node get_first_child_node(com.sun.xml.tree.ElementNode parent,String node_name,int type_wanted)
	throws java.lang.Exception
{
Node child;

child=parent.getFirstChild();					//get 1st child
while (child!=null)						//while child exists
	{
	if (child.getNodeType()==type_wanted)			//if it is the type wanted
		if (node_name==null)				//and if no node_name specified
			return(child);				//found it
		else if (child.getNodeName().equals(node_name))	//otherwise also test if node names equal
			return(child);
	child=child.getNextSibling();				//if fail, try next sibling
	}
return(null);
} //end method get_first_child_node



/**
* Find the next sibling of an <em>ElementNode</em> with a specific type.
*
* @param parent	The parent node, which is an <em>com.sun.xml.tree.ElementNode</em>.
* @throws java.lang.Exception	When an error occurs.
*/
public org.w3c.dom.Node get_next_sibling_node(com.sun.xml.tree.ElementNode current_node,int type_wanted)
	throws java.lang.Exception
{
return(get_next_sibling_node(current_node,null,type_wanted));
} //end method get_next_sibling_node


/**
* Find the next sibling of an <em>com.sun.xml.tree.ElementNode</em> with a specific type,
* matching an optional <code>node_name</code>.
*
* @param parent	The parent node, which is an <em>com.sun.xml.tree.ElementNode</em>.
* @throws java.lang.Exception	When an error occurs.
*/
public org.w3c.dom.Node get_next_sibling_node(com.sun.xml.tree.ElementNode current_node,String node_name,int type_wanted)
	throws java.lang.Exception
{
Node sibling;

sibling=current_node.getNextSibling();				//get sibling of current node
while (sibling!=null)						//if sibling exists
	{
	if (sibling.getNodeType()==type_wanted)			//and the type is the one wanted
		if (node_name==null)
			return(sibling);			//then return it
		else if (sibling.getNodeName().equals(node_name))
			return(sibling);
	sibling=sibling.getNextSibling();			//else loop to next sibling
	}
return(null);
} //end method get_next_sibling_node



/**
* Get the value of the 1st <em>TextNode</em> under a certain <em>ElementNode</em>, start search from the root.
* This is a short-hand method that saves you from fetching the <em>ElementNode</em>, find the <em>TextNode</em> and get its value.
* It gets the first <em>TextNode</em> child of the <em>ElementNode</em> and return it's node value.
* If you have multiple <em>TextNode</em> under the target <em>ElementNode</em>, better use the <code>get_first_child_node</code> and
* <code>get_next_sibling_node methods()</code>.
*
* @param tag_name	The tagname of the <em>ElementNode</em> to search.
*
* @return The value of the element as a <em>String</em>. Or <code>null</code> if cannot find it.
*/
public String get_element_node_text(String tag_name)
{
try	{
	return(get_element_node_text((ElementNode)this.getDocumentElement(),tag_name));	//search from the root node
	} catch (Exception e)
		{
		return(null);
		}
} //end method get_element_node_text

/**
* Get the value of the 1st <em>TextNode</em> after a certain <em>ElementNode</em>.
* This is a short-hand method that saves you from fetching the <em>ElementNode</em>, find the <em>TextNode</em> and get its value.
* It gets the first <em>TextNode</em> child of the <em>ElementNode</em> and return it's node value.
* If you have multiple <em>TextNode</em> under the target <em>ElementNode</em>, better use the <code>get_first_child_node()</code> and
* <code>get_next_sibling_node methods()</code>.
*
* @param parent		The ElementNode to start searching.
* @param tag_name	The tagname of the ElementNode to search.
*
* @return The value of the element as a String. Or null if cannot find it.
*/
public String get_element_node_text(com.sun.xml.tree.ElementNode parent,String tag_name)
	throws java.lang.Exception
{
try	{
	return(get_first_child_node(find_element_node(parent,tag_name),org.w3c.dom.Node.TEXT_NODE).getNodeValue());
	} catch (Exception e)
		{
		return(null);
		}
} //end method get_element_node_text



/**
* Find the <em>sender URI</em> in an <em>AktMessage</em> and return it as a <em>String</em>.
*
* @return	The sender URI as a <em>String</em>.
* @throws	java.net.MalformedURLException	When the returned <em>String</em> is not a valid <em>URL</em>.
* @throws	java.lang.Exception		When other errors occur.
*/
public String get_sender()
	throws	java.net.MalformedURLException,
		java.lang.Exception
{
String uri_string;
URL temp_url;


try	{
	//
	//then find the 1st Text node child
	//and get it's value
	//
	//if anything goes wrong, return null
	//
	//uri_string=get_text_node_child(find_element_node("aktbus:hasSenderURI")).getNodeValue();
	uri_string=get_first_child_node(find_element_node("aktbus:hasSenderURI"),Node.TEXT_NODE).getNodeValue();
	temp_url=new URL(uri_string);				//try to see if the URL is valid
	return(uri_string);
	} catch (Exception e)
		{
		return(null);
		}
}	//end method get_sender


/**
* Find the receiver URI in an <em>AktMessage</em> and return it as a <em>String</em>.
*
* @return	The receiver URI as a <em>String</em>.
* @throws	java.net.MalformedURLException	When the returned <em>String</em> is not a valid <em>URL</em>.
* @throws	java.lang.Exception		When other errors occur.
*/
public String get_receiver()
	throws	java.net.MalformedURLException,
		java.lang.Exception
{
String uri_string;
URL temp_url;

//uri_string=get_text_node_child(find_element_node("aktbus:hasReceiverURI")).getNodeValue();
uri_string=get_first_child_node(find_element_node("aktbus:hasReceiverURI"),Node.TEXT_NODE).getNodeValue();
temp_url=new URL(uri_string);				//try to see if the URL is valid
return(uri_string);
}	//end method get_receiver


/**
* Find the content of the <code>&lt;aktbus:hasSimpleMessage&gt;</code> element.
*
* @return	The value of the element <code>&lt;aktbus:hasSimpleMessage&gt;</code> in a <em>string</em>.
* @throws	java.lang.Exception	When an error occurs.
*/
public String get_simple_message()
	throws	java.lang.Exception
{
return(get_first_child_node(find_element_node("aktbus:hasSimpleMessage"),Node.TEXT_NODE).getNodeValue());
}	//end method get_simple_message


/**
* Send an <em>AktMessage</em> with blocking, using specified receiver URI in the <em>AktMessage</em> object.
*
* @throws	java.lang.Exception		When a general error occurs.
* @throws	java.io.IOException		When I/O-related error occurs.
* @throws	java.io.InterruptedIOException	When a timeout occurs.
*
*/
public void send()
	throws	java.lang.Exception,
		java.io.IOException,
		java.io.InterruptedIOException
{
send(0);
} //end method send


/**
* Send to the receiver as specified in the <em>AktMessage</em>, with timeout.
* A receiver must be specified in the <em>AktMessage</em>.
*
* @param timeout The timeout (in second) on the connection to the receiver.
*
* @throws	java.lang.Exception		When a general error occurs.
* @throws	java.io.IOException		When I/O-related error occurs.
* @throws	java.io.InterruptedIOException	When a timeout occurs.
*/
public void send(int timeout)
	throws	java.lang.Exception,
		java.io.IOException,
		java.io.InterruptedIOException
{
URL receiver_uri;

//
// when you "send", the sender URI must be specified
// or you should use send_and_receive
//
//if (get_sender()==null)				//check if sender URI is specified
//	throw new Exception("No sender URI specified. Better use send_and_receive.");
receiver_uri=new URL(get_receiver());		//get receiver URI
send(receiver_uri, timeout);			//call the send method
} //end method send


/**
* Send to receiver as specified in the <em>AktMessage</em>, with timeout.
*
* @param receiver_uri	The receiver URI that the the message should be sent to.
* @param timeout	The timeout for the socket connection to the receiver.
*
* @throws	java.lang.Exception		When a general error occurs.
* @throws	java.io.IOException		When I/O-related error occurs.
* @throws	java.io.InterruptedIOException	When a timeout occurs.
*/

/* THIS IS THE ORIGINAL SOCKET-BASED VERSION
***
*** COMMENT STARTS
***
public void send(URL receiver_uri, int timeout)
	throws Exception, IOException, InterruptedIOException
{
String receiver_uri_string; //receiver's URI in string
Socket s;
PrintWriter out;
BufferedReader in;
StringWriter sw;
String response_string;		//response from HTTP server
String request_file;		//file/path in POST request



if (receiver_uri.getPort()<0)
	s = new Socket(receiver_uri.getHost(),80);		//create socket & connect to URL
else s = new Socket(receiver_uri.getHost(), receiver_uri.getPort());		//create socket & connect to URL
s.setSoTimeout(timeout);						//set timeout value
in = new BufferedReader(new InputStreamReader(s.getInputStream()));	//create input stream to connection
out = new PrintWriter(s.getOutputStream());				//create output stream

request_file=receiver_uri.getFile();				//get file in receiver's URI
if (request_file.length()==0)					//if file is empty
	request_file="/";					//make it "root"
out.print("POST "+request_file+" HTTP/"+HTTP_VERSION);		//send HTTP request line
out.write(13);							//output HTTP's newline
out.write(10);
sw = new StringWriter();					//create string writer for XML message
ugly_write(sw);							//write XML to buffer
sw.close();							//close string writer
out.print("Content-Length: " + sw.toString().length());		//write content-length to connection
out.write(13);							//end of line + a blank line
out.write(10);
out.write(13);
out.write(10);
out.print(sw.toString());		//now write XML to connection
out.write(13);				//finish with a newline
out.write(10);
out.flush();				//flush output
response_string=in.readLine();		//read reply from receiver
out.close();				//close connections
in.close();
s.close();				//close socket
process_http_response(response_string);	//process response
}
***
*** COMMENT ENDS
***
*/



/**
* Sends an <em>AktMessage</em> to a certain recipient.
* A receiver URI must be specified in the <em>AktMessage</em>, although the two do not
* have to be the same.
*
* @param	sender_uri	The URL of the sender. A <code>null</code> means piggy-packed reply.
* @param	receiver_uri	The URL of the receiver.
* @param	timeout		The timeout value waiting for a response from the other side, in second.
*
* @throws	java.lang.Exception		When sending fails. i.e. reply is not OK.
* @throws	java.io.IOException		When I/O-related error occurs.
* @throws	java.io.InterruptedIOException	When a timeout occurs.
*/
public void send(URL receiver_uri,int timeout)
	throws	java.lang.Exception,
		java.io.IOException,
		java.io.InterruptedIOException
{
PrintWriter out;
BufferedReader in;
StringWriter sw;
HttpRequest req;

//
// make sure that sender URI is specified in the AktMessage
//
//if (get_sender()==null)
//	throw new Exception("No sender URI specified. Better use send_and_receive.");

//
// make sure that receiver URI is specified
//
if (get_receiver()==null)
	throw new Exception("Receiver must be specified for an AktMessage to be sent.");

//
// put message into a String
//
sw=new StringWriter();				//create string writer for XML message
ugly_write(sw);					//ugl-write XML to buffer
sw.close();					//close string writer

//
// Sun's Brazil doesn't put "/" in the HTTP header if it is missing in the end of the URL
// have to put it back by ourselves
//
if (receiver_uri.getFile().length()==0)						//see if file part of URL is empty
	req=new HttpRequest(receiver_uri.toString()+"/");			//if yes, put in "/"
else req=new HttpRequest(receiver_uri);

//
// set parameter values, output POST request
//
req.setMethod("POST");								//set HTTP method to POST
req.setRequestHeader("Content-Type", "text/xml");				//set content-type
out=new PrintWriter(req.getOutputStream());					//create output stream
out.write(sw.toString());							//write string
out.close();									//close output stream

req.connect();									//now connect to server

//
// close input streams, as no piggy-packed reply is expected
//
in = new BufferedReader(new InputStreamReader(req.getInputStream()));		//create input stream
in.close();									//close input stream
req.disconnect();								//disconnect from host
if (req.getResponseCode()!=200)							//examine reponse code
	throw new Exception(req.status);
} //end method send





/**
* Send to receiver and waits for reply from the same connection.
* There must be no sender URI specified in the AktMessage or an Exception will be raised.
*
* @throws	java.net.MalformedURLException	When the receiver URI is not valid.
* @throws	java.io.IOException		When I/O-related error occurs.
* @throws	java.io.InterruptedIOException	When a timeout occurs.
* @throws	java.lang.Exception		When sending fails. i.e. reply is not OK.
*/
public AktMessage send_and_receive()
	throws	java.io.IOException,
		java.io.InterruptedIOException,
		java.net.MalformedURLException,
		java.lang.Exception
{
return(send_and_receive(null,0,null));
}

/**
* Send to receiver and waits for reply from the same connection.
* There must be no sender URI specified in the AktMessage or an Exception will be raised.
*
* @param	receiver_uri	The URL of the receiver.
* @param	timeout		The timeout value waiting for a response from the other side, in second.
* @param	result_class	The class of the expected reply as a java.lang.Class object.
*
* @throws	java.net.MalformedURLException	When the receiver URI is not valid.
* @throws	java.io.IOException		When I/O-related error occurs.
* @throws	java.io.InterruptedIOException	When a timeout occurs.
* @throws	java.lang.Exception		When sending fails. i.e. reply is not OK.
*/
public AktMessage send_and_receive(URL receiver_uri,int timeout,java.lang.Class result_class)
	throws	java.io.IOException,
		java.io.InterruptedIOException,
		java.net.MalformedURLException,
		java.lang.Exception
{
PrintWriter out;		//output stream
BufferedReader in;		//input stream
StringWriter sw;
HttpRequest req;

if (receiver_uri==null)
	{
	receiver_uri=new URL(get_receiver());
	}
if (timeout<0)
	timeout=0;
if (result_class==null)
	result_class=getClass();

//
// make sure that no sender URI is specified in the AktMessage
// or receiver will get confused
//
if (get_sender()!=null)
	throw new Exception("sender URI is NOT null. Better use 'send' instead of 'send_and_receive'.");

//
// put message into a String
//
sw=new StringWriter();				//create string writer for XML message
ugly_write(sw);					//ugl-write XML to buffer
sw.close();					//close string writer

//
// Sun's Brazil doesn't put "/" in the HTTP header if it is missing in the end of the URL
// have to put it back by ourselves
//
if (receiver_uri.getFile().length()==0)						//see if file part of URL is empty
	req=new HttpRequest(receiver_uri.toString()+"/");			//if yes, put in "/"
else req=new HttpRequest(receiver_uri);
//
// set parameter values
//
req.setMethod("POST");								//set HTTP method to POST
req.setRequestHeader("Content-Type", "text/xml");				//set content-type
//
// send POST request to receiver
//
out=new PrintWriter(req.getOutputStream());					//create output stream
out.write(sw.toString());							//write string
out.close();									//close output stream

//
// connect to receiver & wait for piggy-packed reply
//
req.connect();									//now connect to server
in = new BufferedReader(new InputStreamReader(req.getInputStream()));		//create input stream
//
// now receive piggy-packed reply
//
//AktMessage reply=receive_from_server(in);					//try to receive reply
AktMessage reply=receive_from_server(in,result_class);					//try to receive reply
in.close();									//close input stream
req.disconnect();								//disconnect from host
if (req.getResponseCode()!=200)							//examine reponse code
	throw new Exception(req.status);
return(reply);
} //end method send_and_receive


/**
* Receives a String from an input stream an makes it into an AktMessage.
* The String is expected to contain the proper AktMessage header & body.
*
* @param in	The input stream to read from.
* @returns	Returns an AktMessage.
* @throws	java.lang.Exception
* @throws	java.io.IOException
* @throws	java.io.InterruptedIOException
*/
//private AktMessage receive_from_server(BufferedReader in)
private AktMessage receive_from_server(BufferedReader in,java.lang.Class result_class)
	throws	java.lang.Exception,
		java.io.IOException,
		java.io.InterruptedIOException
{
String reply="";
char buffer[]=new char[BUFFER_SIZE];				//buffer for reading reply
int byte_read;

byte_read=in.read(buffer,0,BUFFER_SIZE);			//read 1st chunk
while (byte_read>0)
	{
	reply+=new String(buffer,0,byte_read);			//append to string read
	byte_read=in.read(buffer,0,BUFFER_SIZE);		//read next chunk
	}
//
// invoke constructor of the required class
//
Class[] arg_classes=new Class[1];				//a class array of 1 element
Object[] args=new Object[1];					//an object array of 1 element
Constructor c;							//the constructor to invoke

arg_classes[0]=java.lang.Class.forName("java.lang.String");	//the java.lang.String class object
args[0]=reply;							//put in the arg
c=result_class.getConstructor(arg_classes);			//get the constructor w/ 1 String arg
return((AktMessage)c.newInstance(args));			//call the constructor

//return(new AktMessage(reply));				//create and AktMessage from the String
} //end method receive_from_server



/**
* Ugly-printing an <em>AktMessage</em> to stdout
* without any translation of escape char (entity reference).
* This is useful when you want to print a message without any newline, e.g. into a <em>String</em>.
*
* @throws java.io.IOException	When an I/O error occurs in writing.
*/
public void ugly_write() throws IOException
{
ugly_write(false);
} //end method ugly_write

/**
* Ugly-printing an <em>AktMessage</em> to stdout.
* User can choose to have entity references (like &amp, &gt, &lt) translated.
*
* @param translate	A boolean flag. True if translation wanted, false otherwise.
* @throws java.io.IOException	When an I/O error occurs in writing.
*/
public void ugly_write(boolean translate)
	throws java.io.IOException
{
ugly_write(System.out,translate);
} //end method ugly_write

//
// ugly-print to an OutputStream, using UTF-8 character encoding
//
/**
* Ugly-printing an <em>AktMessage</em> to an <em>OutputStream</em>.
*
* @param out	The <em>OutputStream</em> where the message will be ugly-printed.
* @throws java.io.IOException	When an I/O error occurs in writing.
*/
public void ugly_write(OutputStream out)
	throws java.io.IOException
{
ugly_write(out,false);
} //end method ugly_write


/**
* Ugly-printing an <em>AktMessage</em> to a <em>OutputStream</em>.
*
* @param out		The <em>OutputStream</em> where the message will be ugly-printed.
* @param translate	<code>true</code> to have entity references translated into characters, <code>false</code> otherwise.
* @throws java.io.IOException	When an I/O error occurs in writing.
*/
public void ugly_write(OutputStream out,boolean translate)
	throws java.io.IOException
{
Writer w;

ugly_write(new BufferedWriter(new OutputStreamWriter(out)),"UTF8",translate);	//OutputStream uses UTF-8
} //end method ugly_write


//
// ugly-print to a Writer, using UTF-8 character encoding
//
/**
* Ugly-printing an <em>AktMessage</em> to a <em>Writer</em>.
*
* @param out	The <em>Writer</em> where the message will be ugly-printed.
* @throws java.io.IOException	When an I/O error occurs in writing.
*/
public void ugly_write(Writer out)
	throws java.io.IOException
{
ugly_write(out,false);
} //end method ugly_write


/**
* Ugly-printing an <em>AktMessage</em> to a <em>Writer</em>.
*
* @param out		The <em>Writer</em> where the message will be ugly-printed.
* @param translate	<code>true</code> to have entity references translated into characters, <code>false</code> otherwise.
* @throws java.io.IOException	When an I/O error occurs in writing.
*/
public void ugly_write(Writer out,boolean translate)
	throws java.io.IOException
{
String encoding=null;

if (out instanceof OutputStreamWriter)
	encoding=my_java2std(((OutputStreamWriter)out).getEncoding());
ugly_write(out,encoding,translate);
} //end method ugly_write


/**
* Ugly-printing an <em>AktMessage</em> to a <em>Writer</em>, using a certain encoding.
*
* @param out		The <em>Writer</em> where the message will be ugly-printed.
* @param encoding	The encoding to use.
* @throws java.io.IOException	When an I/O error occurs in writing.
*/
public void ugly_write(Writer out,String encoding)
	throws java.io.IOException
{
ugly_write(out,encoding,false);
} //end method ugly_write


/**
* Ugly-printing an <em>AktMessage</em> to a <em>Writer</em>, using a certain encoding.
*
* @param out		The <em>Writer</em> where the message will be ugly-printed.
* @param encoding	The encoding to use.
* @param translate	<code>true</code> to have entity references translated into characters, <code>false</code> otherwise.
* @throws java.io.IOException	When an I/O error occurs in writing.
*/
public void ugly_write(Writer out,String encoding,boolean translate)
	throws java.io.IOException
{
Node root;

out.write("<?xml version=\"1.0\"");		//write XML header

if (encoding!=null)				//with optional encoding
	{
	out.write(" encoding=\"");
	out.write(encoding);
	out.write("\"");
	}

/*
out.write(" encoding=\"");
out.write("UTF-8");
out.write("\"");
*/

out.write("?>");
ugly_write_tree(out,this.getDocumentElement(),translate); 		//print from root
out.flush();
} //end method ugly_write_xml(void)


//
// dump a tree content to "out" starting from "Node"
//

private void ugly_write_tree(Writer out,Node node)
	throws java.io.IOException
{
ugly_write_tree(out,node,false);
} //end method ugly_write_tree


private void ugly_write_tree(Writer out,Node node,boolean translate)
	throws java.io.IOException
{
Node child;
short node_type;
int i;
NamedNodeMap attributes;

if (node==null)
	return;
node_type=node.getNodeType();	//get node type
switch (node_type)
	{
	case Node.ELEMENT_NODE:		out.write("<"+node.getNodeName());	//write ElementNode tag
					attributes=node.getAttributes();	//get all attribute nodes
					if (attributes!=null)			//and print them all
						{
						for (i=0;i<attributes.getLength();i++)
							{
							ugly_write_tree(out,attributes.item(i),translate);
							}
						}
					out.write(">");
					//
					//now write children
					//
					child=node.getFirstChild();		//get 1st child of this node
					while (child!=null)
						{
						ugly_write_tree(out,child,translate);	//write child & its sub tree
						child=child.getNextSibling();	//try next sibling
						}
					out.write("</"+node.getNodeName()+">");	//write end tag
					break;
	case Node.ATTRIBUTE_NODE:	out.write(" "+node.getNodeName()+"=\""+node.getNodeValue()+"\"");
					break;
	case Node.CDATA_SECTION_NODE:	//out.write(node.getNodeValue());
					escape_write(out,node.getNodeValue());
					break;
	case Node.COMMENT_NODE:		out.write("<!--"+node.getNodeValue()+"-->");
					break;
	case TEXT_NODE:			//out.write(node.getNodeValue());
					escape_write(out,node.getNodeValue());
					break;
	case ENTITY_REFERENCE_NODE:	if (translate)
						out.write(translate_entity_reference(node.getNodeName()));
					else out.write("&"+getNodeName()+";");
					break;

//	*** these seem not to be used for XmlDocument
//	*** prints a warning
//	***
//	*** KH
	default:			out.write("***WARNING:unknown Node type***");
/*
	case DOCUMENT_FRAGMENT_NODE 
				break;
	case DOCUMENT_NODE
				break;
	case DOCUMENT_TYPE_NODE
				break;
	case ENTITY_NODE
				break;
	case ENTITY_REFERENCE_NODE 
				break;
	case NOTATION_NODE 
				break;
	case PROCESSING_INSTRUCTION_NODE 
				break;
*/
		}
} //end method ugly_write_tree(ElementNode)


//
// OK. I have to copy this from XmlDocument as the one in it is private.
//
// K. Hui
//
private String my_java2std(String encodingName)
{
if (encodingName == null)
	return null;

//
// ISO-8859-N is a common family of 8 bit encodings;
// N=1 is the eight bit subset of UNICODE, and there
// seem to be at least drafts for some N >10.
//
if (encodingName.startsWith ("ISO8859_"))	// JDK 1.2
    return "ISO-8859-" + encodingName.substring (8);
if (encodingName.startsWith ("8859_"))		// JDK 1.1
    return "ISO-8859-" + encodingName.substring (5);

// XXX seven bit encodings ISO-2022-* ...
// XXX EBCDIC encodings ...                                                     
if ("ASCII7".equalsIgnoreCase (encodingName)
		|| "ASCII".equalsIgnoreCase (encodingName))
	return "US-ASCII";

//
// All XML parsers _must_ support UTF-8 and UTF-16.
// (UTF-16 ~= ISO-10646-UCS-2 plus surrogate pairs)
//
if ("UTF8".equalsIgnoreCase (encodingName))
	return "UTF-8";
if (encodingName.startsWith ("Unicode"))
	return "UTF-16";

//
// Some common Japanese character sets.
//
if ("SJIS".equalsIgnoreCase (encodingName))
	return "Shift_JIS";                                                   
if ("JIS".equalsIgnoreCase (encodingName))
	return "ISO-2022-JP";
if ("EUCJIS".equalsIgnoreCase (encodingName))
	return "EUC-JP";

// else we can't really do anything
return encodingName;
} //end method my_java2std



/**
* Print out all nodes' type in the XML tree.
* Mainly for debugging purpose.
*
* @param n	The <em>org.w3c.dom.Node</em> from which we start to print in a deep-first manner.
*/
public void print_tree_type(Node n)
{
if (n==null) return;
print_node_type(n);
Node c;

c=n.getFirstChild();
while (c!=null)
	{
	print_tree_type(c);
	c=c.getNextSibling();
	}
}


/**
* Print out the type of a node according to the <em>org.w3c.dom.Node</em> class.
* Mainly for debugging purpose.
*
* @param node The node whose type is to be printed.
*
*/
public static void print_node_type(org.w3c.dom.Node node)
{
switch (node.getNodeType())
	{
	case ATTRIBUTE_NODE:		System.out.println("ATTRIBUTE_NODE");
					break;
	case CDATA_SECTION_NODE:	System.out.println("CDATA_SECTION_NODE");
					break;
	case COMMENT_NODE:		System.out.println("COMMENT_NODE");
					break;
	case DOCUMENT_FRAGMENT_NODE:	System.out.println("DOCUMENT_FRAGMENT_NODE");
					break;
	case DOCUMENT_NODE:		System.out.println("DOCUMENT_NODE");
					break;
	case DOCUMENT_TYPE_NODE:	System.out.println("DOCUMENT_TYPE_NODE");
					break;
	case ELEMENT_NODE:		System.out.println("ELEMENT_NODE");
					break;
	case ENTITY_NODE:		System.out.println("ENTITY_NODE");
					break;
	case ENTITY_REFERENCE_NODE:	System.out.println("ENTITY_REFERENCE_NODE");
					break;
	case NOTATION_NODE:		System.out.println("NOTATION_NODE");
					break;
	case PROCESSING_INSTRUCTION_NODE:System.out.println("PROCESSING_INSTRUCTION_NODE");
					break;
	case TEXT_NODE:			System.out.println("TEXT_NODE");
					break;
	default:	System.out.println("***UNKNOWN***");
	}
}


/**
* Maps Calendar's month value into 1-12
*
* @param calendar_month_value	The integer value from Calendar.MONTH.
* @return	An integer 1-12.
*/
private int map_month(int calendar_month_value)
	throws java.lang.Exception
{
int i;
int map[]={
	Calendar.JANUARY,
	Calendar.FEBRUARY,
	Calendar.MARCH,
	Calendar.APRIL,
	Calendar.MAY,
	Calendar.JUNE,
	Calendar.JULY,
	Calendar.AUGUST,
	Calendar.SEPTEMBER,
	Calendar.OCTOBER,
	Calendar.NOVEMBER,
	Calendar.DECEMBER
	};

for (i=0;i<map.length;i++)
	{
	if (map[i]==calendar_month_value)
		return(i+1);
	}
throw new Exception("Illegal calendar month value");
}	//end method map_month



private void escape_write(Writer out,String to_write)
	throws java.io.IOException
{
int i;

for (i=0;i<to_write.length();i++)
	{
	switch (to_write.charAt(i))
		{
		case '&':	out.write("&amp;");
				break;
		case '>':	out.write("&gt;");
				break;
		case '<':	out.write("&lt;");
				break;
		case '\'':	out.write("&apos;");
				break;
		case '\"':	out.write("&quot;");
				break;
		default:	out.write(to_write.charAt(i));
		}
	}
}

private String translate_entity_reference(String x)
{
if (x==null)
	return(null);
if (x.equalsIgnoreCase("amp"))
	return("&");
else if (x.equalsIgnoreCase("gt"))
	return(">");
else if (x.equalsIgnoreCase("lt"))
	return("<");
else if (x.equalsIgnoreCase("apos"))
	return("'");
else if (x.equalsIgnoreCase("quot"))
	return("\"");
else return("***WARNING: unknown entity reference***");
}



/**
* Initialise the field <code>name_space</code> to contain the <em>HashMap</em> of default XML name space definition.
* When you extend the <em>AktMessage</em>, you may want to add more XML name space definitions.
* You can do so by overriding this method and <code>put</code> more definition into the field <code>name_space</code>,
* which is a <em>HashMap</em>.
* Remember to call the superclass' <code>initialise_namespace()</code> before adding, if you want to include
* the default <em>AktMessage</em> XML name space definition.
*/
protected void initialise_namespace()
{
name_space=new HashMap();
name_space.put("xmlns:rdf","http://www.w3.org/1999/02/22-rdf-syntax-ns#");
name_space.put("xmlns:rdfs","http://www.w3.org/2000/01/rdf-schema#");
name_space.put("xmlns:aktbus","http://www.csd.abdn.ac.uk/~khui/akt/aktbus.xml#");
}
}	//end class AktMessage
