package aktbus.kqml;
import aktbus.core.*;
import com.sun.xml.tree.*;
import org.w3c.dom.*;
import java.net.*;
import java.util.*;

/**
* Representing a KQML message as a subclass of <em>AktMessage</em>.
*
* According to Soton's encoding.
*/
public class KqmlMessage extends AktMessage
{

/**
* Construct a <em>KqmlMessage</em> by parsing a <em>String</em>.
*
* @param body				The <em>String</em> to be parsed.
* @throws org.xml.sax.SAXException	When a parsing error occurs.
* @throws java.lang.Exception		Other general errors.
*/
public KqmlMessage(String body)
	throws org.xml.sax.SAXException, Exception
{
super(body);		//call superclass' constructor to parse it
} //end constructor

/**
* Construct a <em>KqmlMessage</em> from a performative (in a <em>String</em>)
* and an array of <em>KqmlParam</em>.
*
* @param performative				The KQML performative in a <em>String</em>.
* @param param					The KQML parameters in an array of <em>KqmlParam</em>.
* @throws java.net.MalformedURLException	If sender/receiver URL is not valid.
* @throws java.lang.Exception			Other general errors.
*/
public KqmlMessage(String performative,KqmlParam params[])
	throws	java.net.MalformedURLException,
		java.lang.Exception
{
super();				//call superclass' constructor

URL sender=null;
URL receiver=null;

//
//look for sender & receiver in array of KQML parameters
//
int i;

for (i=0;i<params.length;i++)
	{
	if (params[i].name.equals("sender"))
		sender=new URL((String)params[i].value);
	else if (params[i].name.equals("receiver"))
		receiver=new URL((String)params[i].value);
	}
//
// report error if no sender/receiver is specified in the KQML parameters
//
if (sender==null || receiver ==null)
	throw new Exception("sender/receiver not specified in KQML parameters");
Node kqml_content=createNode(performative,params);		//create subtree of KQML content
this.appendChild(createDOM(this,sender,receiver,kqml_content));	//append subtree to this KqmlMessage
initialise_namespace();
} //end constructor


/**
* Create a <em>KqmlMessage</em> from a <em>java.util.Map</em> of <em>KqmlParam</em>.
*
* @param performative	The KQML performative in a <em>String</em>.
* @param params		A <em>java.util.Map</em> of <em>KqmlParam</em>.
* @throws java.net.MalformedURLException	If either sender/receiver URI is not valid.
* @throws java.lang.Exception			When other errors occur.
*/
public KqmlMessage(String performative,Map params)
	throws	java.net.MalformedURLException,
		java.lang.Exception
{
this(performative,params.values());	//get the Map's Collection and call the appropriate constructor
} //end constructor


/**
* Create a <em>KqmlMessage</em> from a <em>java.util.Collection</em> of <em>KqmlParam</em>.
*
* @param performative	The KQML performative in a <em>String</em>.
* @param param		A <em>java.util.Collection</em> of <em>KqmlParam</em>.
*
* @throws java.net.MalformedURLException	When the sender/receiver URL string is not a valid HTTP URL.
* @throws java.lang.Exception			When other general errors occur.
*/
public KqmlMessage(String performative,Collection params)
	throws	java.net.MalformedURLException,
		java.lang.Exception
{
this(performative,(KqmlParam [])(params.toArray(new KqmlParam[0])));	//convert the Collection to an array of KqmlParam and call the appropriate constructor
} //end constructor



/**
* Create s subtree starting with a <code>akt-kqml:KqmlContent</code> element from a <em>Map</em> of <em>KqmlParam</em>.
* Other elements will be added accordingly.
*
* @param performative	The KQML performative in a <code>String</code>.
* @param params		A <em>Map</em> of <em>KqmlParam</em>.
* @returns		The top level <em>Node</em> of the <code>akt-bus:KqmlContent</code> element.
*/
private Node createNode(String performative,Map params)
{
return(createNode(performative,params.values()));		//get params' values as a Collection and proceed
} //end method createNode



/**
* Create s subtree starting with a <code>akt-kqml:KqmlContent</code> element node
* from a <em>Collection</em> of <em>KqmlParam</em>.
*
* @param performative	The KQML performative in a <em>String</em>.
* @param params		A <em>Collection</em> of <em>KqmlParam</em>.
* @returns		The top level <em>Node</em> of <code>akt-bus:KqmlContent</code>.
*/
private Node createNode(String performative,Collection params)
{
return(createNode(performative,(KqmlParam [])params.toArray()));		//get params' values as an Array of KqmlParam
}


/**
* Create s subtree starting with a <code>akt-kqml:KqmlContent</code> element node
* from an array of <em>KqmlParam</em>.
*
* @param performative	The KQML performative in a <em>String</em>.
* @param params		An array of <em>KqmlParam</em>.
* @returns		The top level <em>Node</em> of <code>akt-bus:KqmlContent</code>.
*/
private Node createNode(String performative,KqmlParam params[])
{
ElementNode kc,perf,hp,pm,pn,pv;
int i;

kc=(ElementNode)this.createElement("akt-kqml:KqmlContent");
perf=(ElementNode)this.createElement("akt-kqml:hasPerformative");
perf.appendChild(this.createTextNode(performative));
kc.appendChild(perf);
for (i=0;i<params.length;i++)
	{
	pm=(ElementNode)this.createElement("akt-kqml:"+params[i].name);		//create akt-kqml:* node using KqmlParam's name

	String value_class=params[i].value.getClass().getName();		//get class of param value
	if (value_class.equals("java.lang.String"))				//it is a String
		pm.appendChild(this.createTextNode((String)params[i].value));	//create a TextNode
	else if (value_class.equals("com.sun.xml.Node"))			//if it is a Node
		{
		this.changeNodeOwner((Node)params[i].value);			//change ownership to this XML document
		pm.appendChild((Node)(params[i].value));			//append it to subtree under akt-kqml:hasValue
		}
	else if (value_class.equals("com.sun.xml.tree.XmlDocument"))		//if it is a XmlDocument
		{
		Node root=((XmlDocument)params[i].value).getDocumentElement();	//get root element of XML document
		this.changeNodeOwner(root);					//change ownership to this XML document
		pm.appendChild(root);						//append it to subtree under akt-kqml:hasValue
		}
	else System.out.println("*** unknown class:"+value_class);
	kc.appendChild(pm);						//append KqmlParam node to KqmlContent

	}
return(kc);		//return ElementNode of KqmlContent
} //end method createNode


/**
* Get the performative of a <em>KqmlMessage</em>.
*
* @return	The performative in a <em>String</em>.
*/
public String get_performative()
{
ElementNode kc_node,perf_node;

try	{
	kc_node=find_element_node("akt-kqml:KqmlContent");			//find akt-kqml:KqmlContent node
	return(get_element_node_text(kc_node,"akt-kqml:hasPerformative"));	//return text after akt-kqml:hasPerformative sub-node
	} catch (Exception e)
		{
		return(null);
		}
}


/**
* Returns the set of KQML parameters as a <em>java.util.Map</em> of <em>KqmlParam</em>.
* The returned <em>java.util.Map</em> allows the retrieval of
* <em>KqmlParam</em> (value) by specifying the parameter name (i.e. key).
* e.g. <code>content</code>.
* If you want to iterate through all KQML parameters,
* use the <code>get_parameter_set()</code> method instead.
*
* @returns	The <em>java.util.Map</em> object. Each entry in the <em>java.util.Map</em> is a <em>KqmlParam</em>
*		object indexed by its KQML parameter name (in a <em>String</em>).
*/
public Map get_parameter_map()
{
ElementNode kc_node,param_node;
Node value_node;
String param_name;
HashMap param_map=new HashMap();
KqmlParam km;

try	{
	kc_node=find_element_node("akt-kqml:KqmlContent");		//find akt-kqml:KqmlContent node
	if (kc_node==null)
		throw new Exception("cannot find element akt-kqml:KqmlContent");
	//
	// look for all param under akt-kqml:KqmlContent, except akt-kqml:hasPerformative
	//
	param_node=(ElementNode)get_first_child_node(kc_node,org.w3c.dom.Node.ELEMENT_NODE);	//get 1st child node under akt-kqml:KqmlContent
	while (param_node!=null)
		{
		param_name=param_node.getNodeName();			//get tag name
		if (!param_name.equals("akt-kqml:hasPerformative"))	//if NOT akt-kqml:hasPeformative
			{
			if (param_name.substring(0,9).equals("akt-kqml:"))				//look for "akt-kqml:" prefix
				{
				//
				// depending on class of value node
				// either put the ElementNode as value
				// or use text value
				//
				param_name=param_name.substring(9);					//strip off "akt-kqml:" prefix
				value_node=param_node.getFirstChild();					//get 1st child under param_node
				String value_class=value_node.getClass().getName();			//get class name of value
				if (value_class.equals("com.sun.xml.tree.ElementNode"))
					{
					km=new KqmlParam(param_name,value_node);			//create KqmlParam from ElementNode
					}
				else if (value_class.equals("com.sun.xml.tree.TextNode"))
					{
					km=new KqmlParam(param_name,get_subtree_text(param_node));	//create KqmlParam from value text
					}
				else	{
					throw new Exception("unknown class in KqmlParam value");
					}
				param_map.put(param_name,km);						//put KqmlParam into map
				}
			}
		param_node=(ElementNode)get_next_sibling_node(param_node,org.w3c.dom.Node.ELEMENT_NODE);	//get next node under akt-kqml:KqmlContent
		}
	} catch (Exception e)
		{
		System.err.println(e.toString());
		return(null);
		}
return(param_map);
}



/**
* Return the set of KQML parameters as a <em>java.util.Collection</em> of <em>KqmlParam</em>.
* If you want to get a particular KQML parameter by specifying the KQML parameter name,
* use the <code>get_parameter_map()</code> method instead.
*
* @returns A <em>java.util.Collection</em> object. Each element in the collection is a <em>KqmlParam</em> object.
*/
public Collection get_parameter_set()
{
ElementNode kc_node,param_node;
Node value_node;
String param_name;
HashSet param_set=new HashSet();
KqmlParam km;

try	{
	kc_node=find_element_node("akt-kqml:KqmlContent");		//find akt-kqml:KqmlContent node
	if (kc_node==null)
		throw new Exception("cannot find element akt-kqml:KqmlContent");
	param_node=(ElementNode)get_first_child_node(kc_node,org.w3c.dom.Node.ELEMENT_NODE);	//get 1st child node under akt-kqml:KqmlContent
	while (param_node!=null)
		{
		param_name=param_node.getNodeName();			//get tag name
		if (!param_name.equals("akt-kqml:hasPerformative"))	//if NOT akt-kqml:hasPeformative
			{
			if (param_name.substring(0,9).equals("akt-kqml:"))				//look for "akt-kqml:" prefix
				{
				//
				// depending on class of value node
				// either put the ElementNode as value
				// or use text value
				//
				param_name=param_name.substring(9);					//strip off "akt-kqml:" prefix
				value_node=param_node.getFirstChild();					//get 1st child under param_node
				String value_class=value_node.getClass().getName();			//get class name of value
				if (value_class.equals("com.sun.xml.tree.ElementNode"))
					{
					km=new KqmlParam(param_name,value_node);			//create KqmlParam from ElementNode
					}
				else if (value_class.equals("com.sun.xml.tree.TextNode"))
					{
					km=new KqmlParam(param_name,get_subtree_text(param_node));	//create KqmlParam from value text
					}
				else	{
					throw new Exception("unknown class in KqmlParam value");
					}
				param_set.add(km);						//add new KqmlParam into set
				}
			}
		param_node=(ElementNode)get_next_sibling_node(param_node,org.w3c.dom.Node.ELEMENT_NODE);	//get next node under akt-kqml:KqmlContent
		}
	} catch (Exception e)
		{
		System.err.println(e.toString());
		return(null);
		}
return(param_set);
}

/**
* Initialise name space <em>Map</em>, overridding the one in <em>AktMessage</em>.
*/
protected void initialise_namespace()
{
super.initialise_namespace();		//init field name_space to basic set of XML name space 1st
name_space.put("akt-kqml","http://www.csd.abdn.ac.uk/~khui/akt/akt-kqml#");	//add KQML's XML name space
}

} //end class KqmlMessage
