import com.sun.xml.tree.*;
import org.xml.sax.*;
import org.w3c.dom.*;
import java.util.*;
import java.net.*;
import java.io.*;

/**
 * The OntobrokerResult class, subclasses AKTMessage to create an AKTMessage with
 * the result from an ontobroker query as content.
 *
 * Constructors:
 *
 *	OntobrokerResult(String sender,String receiver, String querystring, String results[]) throws MalformedURLException
 *
 *		Creates an AKTMessage containing the results of a query on an ontobroker
 *      server.
 *		Throws MalformedURLException if either URL is invalid.
 *
 * Object Methods:
 *
 *	void createNode(void)
 *  void setStyle(String stylesheetURI, String type)
 */

public class OntobrokerResult extends AktMessage {

	/**
	 * Constructs an FLogicMessage with the query and String URIs given
	 *
	 * @param String sender The sender of the message.
	 * @param String receiver The receiver of the message.
	 * @param String queryString The query that produced the results
	 * @param String[] results An array of strings containing the results.
	 *
	 * @throws MalformedURLException From AKTMessage
	 */
    public OntobrokerResult(String sender, String receiver, String queryString, String results[]) throws MalformedURLException {
		super();
		Node q = createNode(results, queryString, sender);
		this.appendChild(createDOM(this, new URL(sender), new URL(receiver), q));
	}

	/**
	 * Adds the stylesheet to the document
	 *
	 * @param String stylesheetURI The URI of the stylesheet
	 * @param String type The type of the stylesheet.
	 *
	 */
	public void setStyle(String stylesheetURI, String type) {
		this.insertBefore(createProcessingInstruction("xml-stylesheet", stylesheetURI + " " + type), find_node("rdf:RDF"));
	}

	/**
	 * Creates the XML Representation of the results, following the
	 * ontoresult schema.
	 *
	 * @param String[] results The result set.
	 *
	 * @returns Node The results as XML, to be placed in the AKTMessage as content
	 */
	public Node createNode(String results[], String queryString, String serviceURI) {
		// Boolean array to mark whether this resultset has been dealt with.
		// Represents a limit to the number of queries??
		boolean[] queries = new boolean[100];
		// Current resultset
		int query;
		StringTokenizer st;
		boolean init = false;

		// Create main result node.

		// Create root ontoresult node.
		Node q = this.createElement("onto:OntobrokerQuery");
		// Add the schema reference.
		((ElementNode) q).setAttribute("xmlns:onto", "http://quail.csd.abdn.ac.uk:18080/akt-ontobroker/schema/ontoresult.xml#");
		// Add the query
		Node hq = this.createElement("onto:hasFLogicQuery");
		q.appendChild(hq);
		Node fq = this.createElement("onto:FLogicQuery");
		hq.appendChild(fq);
		Node qv = this.createElement("onto:hasFLogicVars");
		fq.appendChild(qv);
		Node qr = this.createElement("onto:hasFLogicRel");
		fq.appendChild(qr);

		// Add the FLogic query to the message.
		qv.appendChild(this.createTextNode(queryString.substring(0,queryString.indexOf("<-"))));
		qr.appendChild(this.createTextNode(queryString.substring(queryString.indexOf("<-")+2,queryString.length())));

		// And the service
		Node hserv = this.createElement("onto:hasOntobrokerService");
		q.appendChild(hserv);
		hserv.appendChild(this.createTextNode(serviceURI));

		// Begin the results.
		Node r = this.createElement("onto:hasResultSet");
		q.appendChild(r);

		// Create the 1st result set element
		Node rs = this.createElement("onto:ResultSet");
		r.appendChild(rs);

		// Step through array of results
		for (int i = 0; i < results.length; i++) {
			String result = results[i];

			query = Integer.parseInt(result.substring(0,1));
			// If this is a new result...
			if (!queries[query]) {
				// Create a new resultset

				if (init) {
					// Dont do this the 1st time, its already done above
					r = this.createElement("onto:hasResultSet");
					q.appendChild(r);
					rs = this.createElement("onto:ResultSet");
					r.appendChild(rs);
				}
				init = true;

				// With a title
				Node rt = this.createElement("onto:hasResultTitle");
				Node te = this.createElement("onto:ResultTitle");
				rs.appendChild(rt);
				rt.appendChild(te);

				// Where the title elements are given
				st = new StringTokenizer(result, ",");
				while (st.hasMoreTokens()) {
					String tok = st.nextToken();
					Node ec = this.createElement("onto:hasElement");
					ec.appendChild(this.createTextNode(tok.substring(tok.indexOf("\"")+1,tok.lastIndexOf("\""))));
					te.appendChild(ec);
				}
				queries[query] = true;
			} else {
				// Otherwise this is another result for the current resultset
				Node rn = this.createElement("onto:hasResultMember");
				Node re = this.createElement("onto:ResultMember");
				st = new StringTokenizer(result, ",");
				while (st.hasMoreTokens()) {
					String tok = st.nextToken();
					Node ec = this.createElement("onto:hasElement");
					re.appendChild(ec);
					ec.appendChild(this.createTextNode(tok.substring(tok.indexOf("\"")+1,tok.lastIndexOf("\""))));

				}
				rs.appendChild(rn);
				rn.appendChild(re);
			}



		}
		return q;
	}

}