%
% universal socket library front on top of SICStus's socket library
%
% NOTE:	1)	You should not consult this file but
%		the top level "sockets.pl" with automatic
%		machine & platform detection.
%

:- use_module(library(sockets)).		%use SICStus's socket library
:- use_module(library(lists)).

%---------------------------------------------
% create a socket in AF_INET or AF_UNIX domain
%
% a_socket_create(+Domain,-Socket)
%
% Domain=	'AF_INET' for Internet domain socket
%		'AF_UNIX' for UNIX domain socket
% Socket=	socket to be returned
%---------------------------------------------

a_socket_create(Domain,Socket)	:-
	nonvar(Domain),				%make sure that Domain is instantiated
	a_socket_domain(Domain),		%and is 1 of the 2 valid domains
	var(Socket),				%Socket is no instantiated yet
	sockets:socket(Domain,Socket).		%call SICStus' socket lib to create it

%
% the 2 valid domains
%
a_socket_domain('AF_INET').
a_socket_domain('AF_UNIX').



%-----------------
% destroy a socket
% a_socket_destroy(+Socket)
%
% Socket=socket to be destroyed
%-----------------

a_socket_destroy(Socket)	:-
	nonvar(Socket),
	integer(Socket),
	sockets:socket_close(Socket).		%call SICStus to close socket



%-------------------------------
% binds a socket with an address
% a_socket_bind(+Socket,+Address)
%
% Socket=	socket to be binded (created by socket_create/2)
% Address=	address in the form 'AF_INET'(port)
% or		'AF_UNIX'(pathname)
%-------------------------------

a_socket_bind(Socket,'AF_INET'(Port))	:-
	!,
	integer(Socket),
	sockets:current_host(Host),
	sockets:socket_bind(Socket,'AF_INET'(Host,Port)).	%call SICStus


a_socket_bind(Socket,Port)	:-
	integer(Socket),
	sockets:current_host(Host),
	sockets:socket_bind(Socket,'AF_INET'(Host,Port)).	%call SICStus




%----------------------------------------
% listen to a socket and set queue length
% a_socket_listen(+Socket,+Length)
%
% Socket=	socket to listen (created by socket_create/2)
% Length=	integer specifying the queue length
%----------------------------------------
a_socket_listen(Socket,Length)	:-
	integer(Socket),
	integer(Length),
	!,
	sockets:socket_listen(Socket,Length).		%call SICStus





%------------------------------
% accept an incoming connection
% a_socket_accept(+Socket,-Income,-Read,Write)
%
% Socket=	socket for accepting incoming connection
% Income=	returned file-descriptor of incoming connection
% Read=		read stream for incoming connection
% Write=	write stream for incoming connection
%------------------------------
a_socket_accept(Socket,Income,Rw,Rw)	:-
	integer(Socket),
	var(Income),
	var(Rw),
	!,
	sockets:socket_accept(Socket,Income,Rw).	%call SICStus
							%note that SICStus socket is R/W enabled




%---------------------------
% connect to a remote socket
% a_socket_connect(+Socket,+Address,+Read,+Write)
%
% Socket=	socket for making connection (created by socket_create/2)
% Address=	target address for connection in the form 'AF_INET'(host,port)
% or		'AF_UNIX'(pathname) for the UNIX domain
% Read=		read stream to be returned for connection
% Write=	write stream to be returned
%---------------------------
a_socket_connect(Socket,Address,Rw,Rw)	:-
	integer(Socket),
	var(Rw),
	sockets:socket_connect(Socket,Address,Rw).	%call SICStus

%	stream_code(Read,Stream_code),
%	stream_code(Write,Stream_code).



%----------------------------------------------------------------------
% select from lists of file descriptors
% socket_select(+Sockets,+Read_streams,+Timeout,-Ready_connections)
%
% Sockets=	list of sockets for testing read-ready
% Streams=	list of streams for testing read-ready
% Timeout=	time-out value for selection
%		either 0 for no timeout (blocking)
%		or time(Second,Micro-second) for timeout
% Ready_connections=	list of read-ready connection in connection(Socket,Read_stream,Write_stream)
%			where the 3 arguments may not be all instantiated, depending on the
%			support of a particular Prolog system.
%			However, either Socket or Read must be instantiated.
%			In the case of SICStus, the socket no. is not returned when a
%			a connection is made to a server socket.
%			A new read-ready stream may come from a new socket connection
%			or an existing read stream.
%----------------------------------------------------------------------

a_socket_select(Sockets,Read_streams,time(0,0),Ready_connections) :-
	a_socket_select(Sockets,Read_streams,0,Ready_connections).

a_socket_select(Sockets,Read_streams,0,Ready_connections) :-
	socket_select(Sockets,New_streams_from_sockets,off,Read_streams,Ready_streams_from_read),
	findall(connection(_,X,X),member(X,New_streams_from_sockets),Conn_from_sockets),	%get list of connection/3 from socket
	findall(connection(_,Y,Y),member(Y,Ready_streams_from_read),Conn_from_streams),		%get list of connection/3 from read streams
	append(Conn_from_sockets,Conn_from_streams,Ready_connections).

a_socket_select(Sockets,Read_streams,time(Sec,Usec),Ready_connections) :-
	socket_select(Sockets,New_streams_from_sockets,Sec:Usec,Read_streams,Ready_streams_from_read),
	findall(connection(_,X,X),member(X,New_streams_from_sockets),Conn_from_sockets),	%get list of connection/3 from socket
	findall(connection(_,Y,Y),member(Y,Ready_streams_from_read),Conn_from_streams),		%get list of connection/3 from read streams
	append(Conn_from_sockets,Conn_from_streams,Ready_connections).

%----------------------------------------------------------------------
% select from lists of file descriptors
% socket_select(+Reads,+Writes,+Excepts,+Timeout,-Read,-Write,-Except)
%
% Reads=	list of file-descriptors for testing read-ready
% Writes=	list of file-descriptors for testing write-ready
% Excepts=	list of file-descriptors for testing exceptions
% Timeout=	time-out value for selection
%		either 0 for no timeout (blocking)
%		or time(Second,Micro-second) for timeout
% Read=		returned list of file-descriptors ready for reading
% Write=	returned list of file-descriptors ready for writing
% Except=	returned list of file-descriptors with exception
%
% ***NOTE: NOT AVAILABLE IN SICSTUS ANYMORE!!!
%
%----------------------------------------------------------------------

a_socket_select(_,_,_,_,_,_,_)	:-
	format("*** SICStus version doesn't support socket_select/7 anymore!!!***~n",[]),
	fail.

%------------------------------
% get the fully quantified current hostname/address
% hostname(-Host)
%
% Host=	hostname to be returned
%------------------------------
a_hostname(Address)	:-
	sockets:current_host(Local_name),		%call SICStus to get local host name
	sockets:hostname_address(Local_name,Address).	%map into fully quantified address

